@@ -10,7 +10,7 @@
 # distributed under the License is distributed on an "AS IS" BASIS,
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
-# limitations under the License.from __future__ import print_function
+# limitations under the License.
 
 import functools
 import warnings
@@ -25,13 +25,10 @@ from .decors import params_as_tensors_for
 from .quadrature import mvnquad
 from .probability_distributions import Gaussian, DiagonalGaussian, MarkovGaussian
 
-from multipledispatch import dispatch
-from functools import partial
+from .dispatch import dispatch
 
-# By default multipledispatch uses a global namespace in multipledispatch.core.global_namespace
-# We define our own GPflow namespace to avoid any conflict which may arise
-gpflow_md_namespace = dict()
-dispatch = partial(dispatch, namespace=gpflow_md_namespace)
+
+logger = settings.logger()
 
 
 # Sections:
@@ -113,9 +110,6 @@ def _quadrature_expectation(p, obj1, feature1, obj2, feature2, num_gauss_hermite
     """
     num_gauss_hermite_points = 100 if num_gauss_hermite_points is None else num_gauss_hermite_points
 
-    warnings.warn("Quadrature is used to calculate the expectation. This means that "
-                  "an analytical implementations is not available for the given combination.")
-
     if obj2 is None:
         eval_func = lambda x: get_eval_func(obj1, feature1)(x)
     elif obj1 is None:
@@ -155,9 +149,6 @@ def _quadrature_expectation(p, obj1, feature1, obj2, feature2, num_gauss_hermite
     """
     num_gauss_hermite_points = 40 if num_gauss_hermite_points is None else num_gauss_hermite_points
 
-    warnings.warn("Quadrature is used to calculate the expectation. This means that "
-                  "an analytical implementations is not available for the given combination.")
-
     if obj2 is None:
         eval_func = lambda x: get_eval_func(obj1, feature1)(x)
         mu, cov = p.mu[:-1], p.cov[0, :-1]  # cross covariances are not needed
@@ -191,7 +182,7 @@ def expectation(p, obj1, obj2=None, nghp=None):
     :return: a 1-D, 2-D, or 3-D tensor containing the expectation
 
     Allowed combinations
-    
+
     - Psi statistics:
         >>> eKdiag = expectation(p, kern)  (N)  # Psi0
         >>> eKxz = expectation(p, (kern, feat))  (NxM)  # Psi1
@@ -232,8 +223,11 @@ def expectation(p, obj1, obj2=None, nghp=None):
 
     try:
         return _expectation(p, obj1, feat1, obj2, feat2, nghp=nghp)
-    except NotImplementedError as e:
-        print(str(e))
+    except NotImplementedError as e:  # pragma: no cover
+
+        warn_msg = "Quadrature is used to calculate the expectation. " + str(e)
+        logger.warn(warn_msg)
+
         return _quadrature_expectation(p, obj1, feat1, obj2, feat2, nghp)
 
 
@@ -260,7 +254,7 @@ def _expectation(p, kern, feat, none1, none2, nghp=None):
 
     :return: NxM
     """
-    with params_as_tensors_for(kern), params_as_tensors_for(feat):
+    with params_as_tensors_for(kern, feat):
         # use only active dimensions
         Xcov = kern._slice_cov(p.cov)
         Z, Xmu = kern._slice(feat.Z, p.mu)
@@ -300,7 +294,7 @@ def _expectation(p, mean, none, kern, feat, nghp=None):
             message="Currently cannot handle slicing in exKxz.")]):
         Xmu = tf.identity(Xmu)
 
-    with params_as_tensors_for(kern), params_as_tensors_for(feat):
+    with params_as_tensors_for(kern, feat):
         D = tf.shape(Xmu)[1]
         lengthscales = kern.lengthscales if kern.ARD \
             else tf.zeros((D,), dtype=settings.float_type) + kern.lengthscales
@@ -339,7 +333,7 @@ def _expectation(p, mean, none, kern, feat, nghp=None):
             message="Currently cannot handle slicing in exKxz.")]):
         Xmu = tf.identity(Xmu)
 
-    with params_as_tensors_for(kern), params_as_tensors_for(feat):
+    with params_as_tensors_for(kern, feat):
         D = tf.shape(Xmu)[1]
         lengthscales = kern.lengthscales if kern.ARD \
             else tf.zeros((D,), dtype=settings.float_type) + kern.lengthscales
@@ -367,58 +361,109 @@ def _expectation(p, kern1, feat1, kern2, feat2, nghp=None):
     Compute the expectation:
     expectation[n] = <Ka_{Z1, x_n} Kb_{x_n, Z2}>_p(x_n)
         - Ka_{.,.}, Kb_{.,.} :: RBF kernels
-    Ka and Kb as well as Z1 and Z2 can differ from each other, but this is supported
-    only if the Gaussian p is Diagonal (p.cov NxD) and Ka, Kb have disjoint active_dims
-    in which case the joint expectations simplify into a product of expectations
+    Ka and Kb as well as Z1 and Z2 can differ from each other.
 
-    :return: NxMxM
+    :return: N x dim(Z1) x dim(Z2)
     """
-    if kern1.on_separate_dims(kern2) and isinstance(p, DiagonalGaussian):  # no joint expectations required
+
+    if kern1.on_separate_dims(kern2) and isinstance(p, DiagonalGaussian):
+        # no joint expectations required
         eKxz1 = expectation(p, (kern1, feat1))
         eKxz2 = expectation(p, (kern2, feat2))
         return eKxz1[:, :, None] * eKxz2[:, None, :]
 
-    if feat1 != feat2 or kern1 != kern2:
-        raise NotImplementedError("The expectation over two kernels has only an "
-                                  "analytical implementation if both kernels are equal.")
+    Ka, Kb = kern1, kern2
 
-    kern = kern1
-    feat = feat1
+    with params_as_tensors_for(Ka, feat1, Kb, feat2):
 
-    with params_as_tensors_for(kern), params_as_tensors_for(feat):
         # use only active dimensions
-        Xcov = kern._slice_cov(tf.matrix_diag(p.cov) if isinstance(p, DiagonalGaussian) else p.cov)
-        Z, Xmu = kern._slice(feat.Z, p.mu)
+        Xcov = Ka._slice_cov(tf.matrix_diag(p.cov) if isinstance(p, DiagonalGaussian) else p.cov)
+
+        Z1, Xmu = Ka._slice(feat1.Z, p.mu)
 
         N = tf.shape(Xmu)[0]
         D = tf.shape(Xmu)[1]
 
-        squared_lengthscales = kern.lengthscales ** 2. if kern.ARD \
-            else tf.zeros((D,), dtype=settings.tf_float) + kern.lengthscales ** 2.
+        def get_squared_length_scales(kern):
+            squared_lengthscales = kern.lengthscales ** 2. if kern.ARD \
+                        else tf.zeros((D,), dtype=settings.tf_float) + kern.lengthscales ** 2.
+            return squared_lengthscales
 
-        sqrt_det_L = tf.reduce_prod(0.5 * squared_lengthscales) ** 0.5
-        C = tf.cholesky(0.5 * tf.matrix_diag(squared_lengthscales) + Xcov)  # NxDxD
+        if Ka == Kb:
+            La = get_squared_length_scales(Ka)
+            Lb = La
+            half_mean_L = La * 0.5  # average length scale
+        else:
+            La, Lb = map(get_squared_length_scales, (Ka, Kb))
+            half_mean_L = La * Lb / (La + Lb)  # average length scale
+
+        sqrt_det_L = tf.reduce_prod(half_mean_L) ** 0.5
+        C = tf.cholesky(tf.matrix_diag(half_mean_L) + Xcov)  # [N, D, D]
         dets = sqrt_det_L / tf.exp(tf.reduce_sum(tf.log(tf.matrix_diag_part(C)), axis=1))  # N
 
-        C_inv_mu = tf.matrix_triangular_solve(C, tf.expand_dims(Xmu, 2), lower=True)  # NxDx1
-        C_inv_z = tf.matrix_triangular_solve(C,
-                                             tf.tile(tf.expand_dims(tf.transpose(Z) / 2., 0), [N, 1, 1]),
-                                             lower=True)  # NxDxM
-        mu_CC_inv_mu = tf.expand_dims(tf.reduce_sum(tf.square(C_inv_mu), 1), 2)  # Nx1x1
-        z_CC_inv_z = tf.reduce_sum(tf.square(C_inv_z), 1)  # NxM
-        zm_CC_inv_zn = tf.matmul(C_inv_z, C_inv_z, transpose_a=True)  # NxMxM
-        two_z_CC_inv_mu = 2 * tf.matmul(C_inv_z, C_inv_mu, transpose_a=True)[:, :, 0]  # NxM
+        # for mahalanobis computation we need Zᵀ (CCᵀ)⁻¹ Z  as well as C⁻¹ Z
+        # with Z = Z₁, Z₂  for two rbf kernels
+        def get_cholesky_solve_terms(Z, C=C):
+            C_inv_z = tf.matrix_triangular_solve(
+                C, tf.tile(tf.expand_dims(tf.transpose(Z), 0),
+                           [N, 1, 1]), lower=True)  # [N, D, M]
+            z_CC_inv_z = tf.reduce_sum(tf.square(C_inv_z), 1)  # [N, M]
 
-        exponent_mahalanobis = mu_CC_inv_mu + tf.expand_dims(z_CC_inv_z, 1) + \
-                               tf.expand_dims(z_CC_inv_z, 2) + 2 * zm_CC_inv_zn - \
-                               tf.expand_dims(two_z_CC_inv_mu, 2) - tf.expand_dims(two_z_CC_inv_mu, 1)  # NxMxM
-        exponent_mahalanobis = tf.exp(-0.5 * exponent_mahalanobis)  # NxMxM
+            return C_inv_z, z_CC_inv_z
 
-        # Compute sqrt(self.K(Z)) explicitly to prevent automatic gradient from
-        # being NaN sometimes, see pull request #615
-        kernel_sqrt = tf.exp(-0.25 * kern.square_dist(Z, None))
-        return kern.variance ** 2 * kernel_sqrt * \
-               tf.reshape(dets, [N, 1, 1]) * exponent_mahalanobis
+        C_inv_mu = tf.matrix_triangular_solve(C, tf.expand_dims(Xmu, 2), lower=True)  # [N, D, 1]
+        mu_CC_inv_mu = tf.expand_dims(tf.reduce_sum(tf.square(C_inv_mu), 1), 2)  # [N, 1, 1]
+
+        C_inv_z1, z1_CC_inv_z1 = get_cholesky_solve_terms(Z1 / La * half_mean_L)
+        z1_CC_inv_mu = 2 * tf.matmul(C_inv_z1, C_inv_mu, transpose_a=True)[:, :, 0]  # [N, M1]
+
+        if feat1 == feat2 and Ka == Kb:
+            # in this case Z2==Z1 so we can reuse the Z1 terms
+            C_inv_z2, z2_CC_inv_z2  = C_inv_z1, z1_CC_inv_z1
+            z2_CC_inv_mu = z1_CC_inv_mu  # [N, M]
+            Z2 = Z1
+        else:
+            # compute terms related to Z2
+            Z2, _ = Kb._slice(feat2.Z, p.mu)
+            C_inv_z2, z2_CC_inv_z2 = get_cholesky_solve_terms(Z2 / Lb * half_mean_L)
+            z2_CC_inv_mu = 2 * tf.matmul(C_inv_z2, C_inv_mu, transpose_a=True)[:, :, 0]  # [N, M2]
+
+        z1_CC_inv_z2 = tf.matmul(C_inv_z1, C_inv_z2, transpose_a=True)  # [N, M1, M2]
+
+        # expand dims for broadcasting
+        # along M1
+        z2_CC_inv_mu = tf.expand_dims(z2_CC_inv_mu, 1)  # [N, 1, M2]
+        z2_CC_inv_z2 = tf.expand_dims(z2_CC_inv_z2, 1)
+
+        # along M2
+        z1_CC_inv_mu = tf.expand_dims(z1_CC_inv_mu, 2)  # [N, M1, 1]
+        z1_CC_inv_z1 = tf.expand_dims(z1_CC_inv_z1, 2)
+
+        # expanded version of ((Z1 + Z2)-mu) (CCT)-1 ((Z1 + Z2)-mu)
+        mahalanobis = mu_CC_inv_mu + z2_CC_inv_z2 + \
+                      z1_CC_inv_z1 + 2 * z1_CC_inv_z2 - \
+                      z1_CC_inv_mu - z2_CC_inv_mu  # [N, M1, M2]
+
+        exp_mahalanobis = tf.exp(-0.5 * mahalanobis)  # [N, M1, M2]
+
+        if Z1 == Z2:
+            # CAVEAT : Compute sqrt(self.K(Z)) explicitly
+            # to prevent automatic gradient from
+            # being NaN sometimes, see pull request #615
+            sqrt_exp_dist = tf.exp(-0.25 * Ka.scaled_square_dist(Z1, None))
+        else:
+            # Compute exp( -.5 (Z-Z')^top (L_1+L_2)^{-1} (Z-Z') )
+            lengthscales_rms = tf.sqrt(La + Lb)
+            Z1 = Z1 / lengthscales_rms
+            Z1sqr = tf.reduce_sum(tf.square(Z1), axis=1)
+            Z2 = Z2 / lengthscales_rms
+            Z2sqr = tf.reduce_sum(tf.square(Z2), axis=1)
+            dist = -2 * tf.matmul(Z1, Z2, transpose_b=True) \
+                   + tf.reshape(Z1sqr, (-1, 1)) + tf.reshape(Z2sqr, (1, -1))
+            sqrt_exp_dist = tf.exp(-0.5 * dist)  # M1 x M2
+
+        return Ka.variance * Kb.variance * sqrt_exp_dist * \
+               tf.reshape(dets, [N, 1, 1]) * exp_mahalanobis
 
 
 # =============================== Linear Kernel ===============================
@@ -449,7 +494,7 @@ def _expectation(p, kern, feat, none1, none2, nghp=None):
 
     :return: NxM
     """
-    with params_as_tensors_for(kern), params_as_tensors_for(feat):
+    with params_as_tensors_for(kern, feat):
         # use only active dimensions
         Z, Xmu = kern._slice(feat.Z, p.mu)
 
@@ -472,7 +517,7 @@ def _expectation(p, kern, feat, mean, none, nghp=None):
             message="Currently cannot handle slicing in exKxz.")]):
         Xmu = tf.identity(Xmu)
 
-    with params_as_tensors_for(kern), params_as_tensors_for(feat):
+    with params_as_tensors_for(kern, feat):
         N = tf.shape(Xmu)[0]
         var_Z = kern.variance * feat.Z  # MxD
         tiled_Z = tf.tile(tf.expand_dims(var_Z, 0), (N, 1, 1))  # NxMxD
@@ -496,7 +541,7 @@ def _expectation(p, kern, feat, mean, none, nghp=None):
             message="Currently cannot handle slicing in exKxz.")]):
         Xmu = tf.identity(Xmu)
 
-    with params_as_tensors_for(kern), params_as_tensors_for(feat):
+    with params_as_tensors_for(kern, feat):
         N = tf.shape(Xmu)[0] - 1
         var_Z = kern.variance * feat.Z  # MxD
         tiled_Z = tf.tile(tf.expand_dims(var_Z, 0), (N, 1, 1))  # NxMxD
@@ -528,7 +573,7 @@ def _expectation(p, kern1, feat1, kern2, feat2, nghp=None):
     kern = kern1
     feat = feat1
 
-    with params_as_tensors_for(kern), params_as_tensors_for(feat):
+    with params_as_tensors_for(kern, feat):
         # use only active dimensions
         Xcov = kern._slice_cov(tf.matrix_diag(p.cov) if isinstance(p, DiagonalGaussian) else p.cov)
         Z, Xmu = kern._slice(feat.Z, p.mu)
@@ -743,7 +788,7 @@ def _expectation(p, mean1, none1, mean2, none2, nghp=None):
 
     :return: NxQ1xQ2
     """
-    with params_as_tensors_for(mean1), params_as_tensors_for(mean2):
+    with params_as_tensors_for(mean1, mean2):
         e_xxt = p.cov + (p.mu[:, :, None] * p.mu[:, None, :])  # NxDxD
         e_A1t_xxt_A2 = tf.einsum("iq,nij,jz->nqz", mean1.A, e_xxt, mean2.A)  # NxQ1xQ2
         e_A1t_x_b2t = tf.einsum("iq,ni,z->nqz", mean1.A, p.mu, mean2.b)  # NxQ1xQ2
@@ -765,7 +810,7 @@ def _expectation(p, kern, none1, none2, none3, nghp=None):
     :return: N
     """
     return functools.reduce(tf.add, [
-        expectation(p, k, nghp=nghp) for k in kern.kern_list])
+        expectation(p, k, nghp=nghp) for k in kern.kernels])
 
 
 @dispatch(Gaussian, kernels.Sum, InducingPoints, type(None), type(None))
@@ -778,7 +823,7 @@ def _expectation(p, kern, feat, none2, none3, nghp=None):
     :return: NxM
     """
     return functools.reduce(tf.add, [
-        expectation(p, (k, feat), nghp=nghp) for k in kern.kern_list])
+        expectation(p, (k, feat), nghp=nghp) for k in kern.kernels])
 
 
 @dispatch(Gaussian,
@@ -793,7 +838,7 @@ def _expectation(p, mean, none, kern, feat, nghp=None):
     :return: NxQxM
     """
     return functools.reduce(tf.add, [
-        expectation(p, mean, (k, feat), nghp=nghp) for k in kern.kern_list])
+        expectation(p, mean, (k, feat), nghp=nghp) for k in kern.kernels])
 
 
 @dispatch(MarkovGaussian, mean_functions.Identity, type(None), kernels.Sum, InducingPoints)
@@ -806,7 +851,7 @@ def _expectation(p, mean, none, kern, feat, nghp=None):
     :return: NxDxM
     """
     return functools.reduce(tf.add, [
-        expectation(p, mean, (k, feat), nghp=nghp) for k in kern.kern_list])
+        expectation(p, mean, (k, feat), nghp=nghp) for k in kern.kernels])
 
 
 @dispatch((Gaussian, DiagonalGaussian), kernels.Sum, InducingPoints, kernels.Sum, InducingPoints)
@@ -821,15 +866,15 @@ def _expectation(p, kern1, feat1, kern2, feat2, nghp=None):
     crossexps = []
 
     if kern1 == kern2 and feat1 == feat2:  # avoid duplicate computation by using transposes
-        for i, k1 in enumerate(kern1.kern_list):
+        for i, k1 in enumerate(kern1.kernels):
             crossexps.append(expectation(p, (k1, feat1), (k1, feat1), nghp=nghp))
 
-            for k2 in kern1.kern_list[:i]:
+            for k2 in kern1.kernels[:i]:
                 eKK = expectation(p, (k1, feat1), (k2, feat2), nghp=nghp)
                 eKK += tf.matrix_transpose(eKK)
                 crossexps.append(eKK)
     else:
-        for k1, k2 in it.product(kern1.kern_list, kern2.kern_list):
+        for k1, k2 in it.product(kern1.kernels, kern2.kernels):
             crossexps.append(expectation(p, (k1, feat1), (k2, feat2), nghp=nghp))
 
     return functools.reduce(tf.add, crossexps)
@@ -860,8 +905,7 @@ def _expectation(p, rbf_kern, feat1, lin_kern, feat2, nghp=None):
     if rbf_kern.active_dims != lin_kern.active_dims:
         raise NotImplementedError("active_dims have to be the same for both kernels.")
 
-    with params_as_tensors_for(rbf_kern), params_as_tensors_for(lin_kern), \
-         params_as_tensors_for(feat1), params_as_tensors_for(feat2):
+    with params_as_tensors_for(rbf_kern, lin_kern, feat1, feat2):
         # use only active dimensions
         Xcov = rbf_kern._slice_cov(tf.matrix_diag(p.cov) if isinstance(p, DiagonalGaussian) else p.cov)
         Z, Xmu = rbf_kern._slice(feat1.Z, p.mu)
@@ -914,7 +958,7 @@ def _expectation(p, lin_kern, feat1, rbf_kern, feat2, nghp=None):
 
 
 # ============================== Product kernels ==============================
-# Note: product kernels are only supported if the kernels in kern.kern_list act
+# Note: product kernels are only supported if the kernels in kern.kernels act
 #       on disjoint sets of active_dims and the Gaussian we are integrating over
 #       is Diagonal
 
@@ -933,7 +977,7 @@ def _expectation(p, kern, none1, none2, none3, nghp=None):
             "Product currently needs to be defined on separate dimensions.")  # pragma: no cover
 
     return functools.reduce(tf.multiply, [
-        expectation(p, k, nghp=nghp) for k in kern.kern_list])
+        expectation(p, k, nghp=nghp) for k in kern.kernels])
 
 
 @dispatch(DiagonalGaussian, kernels.Product, InducingPoints, type(None), type(None))
@@ -951,7 +995,7 @@ def _expectation(p, kern, feat, none2, none3, nghp=None):
             "Product currently needs to be defined on separate dimensions.")  # pragma: no cover
 
     return functools.reduce(tf.multiply, [
-        expectation(p, (k, feat), nghp=nghp) for k in kern.kern_list])
+        expectation(p, (k, feat), nghp=nghp) for k in kern.kernels])
 
 
 @dispatch(DiagonalGaussian, kernels.Product, InducingPoints, kernels.Product, InducingPoints)
@@ -981,7 +1025,7 @@ def _expectation(p, kern1, feat1, kern2, feat2, nghp=None):
             "Product currently needs to be defined on separate dimensions.")  # pragma: no cover
 
     return functools.reduce(tf.multiply, [
-        expectation(p, (k, feat), (k, feat), nghp=nghp) for k in kern.kern_list])
+        expectation(p, (k, feat), (k, feat), nghp=nghp) for k in kern.kernels])
 
 
 # ============== Conversion to Gaussian from Diagonal or Markov ===============
