@@ -246,6 +246,14 @@ class Parameter(Node):
             return Build.NO
         return Build.YES
 
+    @property
+    def trainable(self):
+        return self._trainable
+
+    @trainable.setter
+    def trainable(self, value):
+        self.set_trainable(value)
+
     def set_trainable(self, value):
         if not isinstance(value, bool):
             raise ValueError('Fixed property value must be boolean.')
@@ -263,7 +271,7 @@ class Parameter(Node):
             else:
                 misc.remove_from_trainables(self.parameter_tensor, graph)
 
-        object.__setattr__(self, 'trainable', value)
+        self._trainable = value
 
     def assign(self, value, session=None, dtype=None, force=True):
         if self._externally_defined:
@@ -293,11 +301,21 @@ class Parameter(Node):
     def as_pandas_table(self):
         column_names = ['class', 'prior', 'transform', 'trainable', 'shape', 'fixed_shape', 'value']
         column_values = [self.__class__.__name__, str(self.prior), str(self.transform),
-                         self.trainable, self.shape, self.fixed_shape, self.value]
+                         self.trainable, self.shape, self.fixed_shape, self.value.copy()]
         column_values = [[value] for value in column_values]
         df = misc.pretty_pandas_table([self.pathname], column_names, column_values)
         return df
 
+    def tf_compilation_index(self):
+        """
+        Takes out index from the parameter's tensor name. E.g. parameter tensor name is 
+        GPR-0000/kern/lengthscales, the method for that parameter will return '0000' index.
+        """
+        if self.parameter_tensor is None:
+            return None
+        name = self.parameter_tensor.name
+        return name.split('-', 1)[-1].split('/')[0]
+
     def _valid_input(self, value, dtype=None):
         if not misc.is_valid_param_value(value):
             msg = 'The value must be either a tensorflow variable, an array or a scalar.'
@@ -472,6 +490,9 @@ class Parameter(Node):
     def __str__(self):
         return str(self.as_pandas_table())
 
+    def _repr_html_(self):
+        return self.as_pandas_table()._repr_html_()
+
     @property
     def fixed(self):
         raise NotImplementedError("`fixed` property is no longer supported. Please use `trainable` instead.")
