@@ -10,36 +10,34 @@
 # distributed under the License is distributed on an "AS IS" BASIS,
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
-# limitations under the License.from __future__ import print_function
-
-import six
-import tensorflow as tf
+# limitations under the License.
 
 import numpy as np
+import tensorflow as tf
 from numpy.testing import assert_allclose
 
 import gpflow
 from gpflow import settings
 from gpflow.test_util import GPflowTestCase
 
-import pytest
-
 
 class LikelihoodSetup(object):
     def __init__(self, likelihood, Y, tolerance):
         self.likelihood, self.Y, self.tolerance = likelihood, Y, tolerance
-        self.is_analytic = six.get_unbound_function(likelihood.predict_density) is not\
-            six.get_unbound_function(gpflow.likelihoods.Likelihood.predict_density)
+        self.is_analytic = likelihood.predict_density is not \
+                           gpflow.likelihoods.Likelihood.predict_density
 
 
 def getLikelihoodSetups(includeMultiClass=True, addNonStandardLinks=False):
     test_setups = []
     rng = np.random.RandomState(1)
     for likelihoodClass in gpflow.likelihoods.Likelihood.__subclasses__():
+        if likelihoodClass == gpflow.likelihoods.MonteCarloLikelihood:
+            continue  # abstract base class
         if likelihoodClass == gpflow.likelihoods.Ordinal:
             test_setups.append(
                 LikelihoodSetup(likelihoodClass(np.array([-1, 1])),
-                          rng.randint(0, 3, (10, 2)), 1e-6))
+                                rng.randint(0, 3, (10, 2)), 1e-6))
         elif likelihoodClass == gpflow.likelihoods.SwitchedLikelihood:
             continue  # switched likelihood tested separately
         elif likelihoodClass == gpflow.likelihoods.MultiClass:
@@ -49,25 +47,26 @@ def getLikelihoodSetups(includeMultiClass=True, addNonStandardLinks=False):
                 tolerance = 1e-3
                 test_setups.append(
                     LikelihoodSetup(likelihoodClass(2),
-                              np.argmax(sample, 1).reshape(-1, 1), tolerance))
+                                    np.argmax(sample, 1).reshape(-1, 1), tolerance))
         else:
             # most likelihoods follow this standard:
             test_setups.append(
                 LikelihoodSetup(likelihoodClass(),
-                          rng.rand(10, 2).astype(settings.float_type), 1e-6))
+                                rng.rand(10, 2).astype(settings.float_type), 1e-6))
 
     if addNonStandardLinks:
         test_setups.append(LikelihoodSetup(gpflow.likelihoods.Poisson(invlink=tf.square),
-                                     rng.rand(10, 2).astype(settings.float_type), 1e-6))
+                                           rng.rand(10, 2).astype(settings.float_type), 1e-6))
         test_setups.append(LikelihoodSetup(gpflow.likelihoods.Exponential(invlink=tf.square),
-                                     rng.rand(10, 2).astype(settings.float_type), 1e-6))
+                                           rng.rand(10, 2).astype(settings.float_type), 1e-6))
         test_setups.append(LikelihoodSetup(gpflow.likelihoods.Gamma(invlink=tf.square),
-                                     rng.rand(10, 2).astype(settings.float_type), 1e-6))
+                                           rng.rand(10, 2).astype(settings.float_type), 1e-6))
 
         def sigmoid(x):
-            return 1./(1 + tf.exp(-x))
+            return 1. / (1 + tf.exp(-x))
+
         test_setups.append(LikelihoodSetup(gpflow.likelihoods.Bernoulli(invlink=sigmoid),
-                                     rng.rand(10, 2).astype(settings.float_type), 1e-6))
+                                           rng.rand(10, 2).astype(settings.float_type), 1e-6))
     return test_setups
 
 
@@ -77,6 +76,7 @@ class TestPredictConditional(GPflowTestCase):
     give the same result as the predict_mean_and_var function if the prediction
     has no uncertainty.
     """
+
     def setUp(self):
         self.test_graph = tf.Graph()
 
@@ -130,9 +130,10 @@ class TestPredictConditional(GPflowTestCase):
 
 class TestQuadrature(GPflowTestCase):
     """
-    Where quadratre methods have been overwritten, make sure the new code
-     does something close to the quadrature
+    Where quadrature methods have been overwritten, make sure the new code
+    does something close to the quadrature
     """
+
     def setUp(self):
         self.test_graph = tf.Graph()
         self.rng = np.random.RandomState()
@@ -157,10 +158,10 @@ class TestQuadrature(GPflowTestCase):
                 # compile and run the functions:
                 F1 = session.run(F1)
                 F2 = session.run(F2)
-                self.assertTrue(np.allclose(F1, F2, test_setup.tolerance, test_setup.tolerance))
+                assert_allclose(F1, F2, test_setup.tolerance, test_setup.tolerance)
 
     def test_pred_density(self):
-        # get all the likelihoods where predict_density  has been overwritten.
+        # get all the likelihoods where predict_density has been overwritten.
         for test_setup in self.test_setups:
             with self.test_context() as session:
                 if not test_setup.is_analytic:
@@ -174,13 +175,159 @@ class TestQuadrature(GPflowTestCase):
                 # compile and run the functions:
                 F1 = session.run(F1)
                 F2 = session.run(F2)
-                self.assertTrue(np.allclose(F1, F2, test_setup.tolerance, test_setup.tolerance))
+                assert_allclose(F1, F2, test_setup.tolerance, test_setup.tolerance)
+
+    def test_pred_mean_and_var(self):
+        # get all the likelihoods where predict_density has been overwritten.
+        for test_setup in self.test_setups:
+            with self.test_context() as session:
+                if not test_setup.is_analytic:
+                    continue
+                l = test_setup.likelihood
+                l.compile()
+                # 'build' the functions
+                F1 = l.predict_mean_and_var(self.Fmu, self.Fvar)
+                F2 = gpflow.likelihoods.Likelihood.predict_mean_and_var(l, self.Fmu, self.Fvar)
+                # compile and run the functions:
+                F1 = session.run(F1)
+                F2 = session.run(F2)
+                assert_allclose(F1, F2, test_setup.tolerance, test_setup.tolerance)
+
+
+class TestMonteCarlo(GPflowTestCase):
+    def setUp(self):
+        self.test_graph = tf.Graph()
+        self.rng = np.random.RandomState()
+        self.rng.seed(1)
+        self.Fmu, self.Fvar, self.Y = self.rng.randn(3, 10, 1).astype(settings.float_type)
+        self.Fvar = 0.01 * (self.Fvar ** 2)
+
+    def test_var_exp(self):
+        with self.test_context() as session:
+            tf.set_random_seed(1)
+            l = gpflow.likelihoods.GaussianMC(0.3)
+            l.num_monte_carlo_points = 1000000
+            # 'build' the functions
+            l.compile()
+            F1 = l.variational_expectations(self.Fmu, self.Fvar, self.Y)
+            F2 = gpflow.likelihoods.Gaussian.variational_expectations(
+                l, self.Fmu, self.Fvar, self.Y)
+            # compile and run the functions:
+            F1 = session.run(F1)
+            F2 = session.run(F2)
+            assert_allclose(F1, F2, rtol=5e-4, atol=1e-4)
+
+    def test_pred_density(self):
+        with self.test_context() as session:
+            tf.set_random_seed(1)
+            l = gpflow.likelihoods.GaussianMC(0.3)
+            l.num_monte_carlo_points = 1000000
+            l.compile()
+            # 'build' the functions
+            F1 = l.predict_density(self.Fmu, self.Fvar, self.Y)
+            F2 = gpflow.likelihoods.Gaussian.predict_density(l, self.Fmu, self.Fvar, self.Y)
+            # compile and run the functions:
+            F1 = session.run(F1)
+            F2 = session.run(F2)
+            assert_allclose(F1, F2, rtol=5e-4, atol=1e-4)
+
+    def test_pred_mean_and_var(self):
+        with self.test_context() as session:
+            tf.set_random_seed(1)
+            l = gpflow.likelihoods.GaussianMC(0.3)
+            l.num_monte_carlo_points = 1000000
+            l.compile()
+            # 'build' the functions
+            F1 = l.predict_mean_and_var(self.Fmu, self.Fvar)
+            F2 = gpflow.likelihoods.Gaussian.predict_mean_and_var(l, self.Fmu, self.Fvar)
+            # compile and run the functions:
+            F1m, F1v = session.run(F1)
+            F2m, F2v = session.run(F2)
+            assert_allclose(F1m, F2m, rtol=5e-4, atol=1e-4)
+            assert_allclose(F1v, F2v, rtol=5e-4, atol=1e-4)
+
+
+class TestSoftMax(GPflowTestCase):
+    def setUp(self):
+        self.test_graph = tf.Graph()
+        self.rng = np.random.RandomState(1)
+
+    def prepare(self, dimF, dimY, num=10):
+        feed = {}
+
+        def make_tensor(data, dtype=settings.float_type):
+            tensor = tf.placeholder(dtype)
+            feed[tensor] = data.astype(dtype)
+            return tensor
+
+        dF = np.vstack((self.rng.randn(num - 3, dimF), np.array([[-3., 0.], [3, 0.], [0., 0.]]))) if dimF == 2 else \
+            self.rng.randn(num, dimF)
+        dY = np.vstack((self.rng.randn(num - 3, dimY), np.ones((3, dimY)))) > 0
+        F = make_tensor(dF)
+        Y = make_tensor(dY, settings.int_type)  # 0 or 1
+        return F, Y, feed
+
+    def test_y_shape_assert(self):
+        """
+        SoftMax assumes the class is given as a label (not, e.g., one-hot
+        encoded), and hence just uses the first column of Y. To prevent
+        silent errors, there is a tf assertion that ensures Y only has one
+        dimension. This test checks that this assert works as intended.
+        """
+        with self.test_context() as sess:
+            F, Y, feed = self.prepare(dimF=5, dimY=2)
+            l = gpflow.likelihoods.SoftMax(5)
+            l.compile()
+            try:
+                sess.run(l.logp(F, Y), feed_dict=feed)
+            except tf.errors.InvalidArgumentError as e:
+                assert "assertion failed" in e.message
+
+    def test_bernoulli_equivalence(self):
+        with self.test_context() as sess:
+            F, Y, feed = self.prepare(dimF=2, dimY=1)
+            Fvar = tf.exp(tf.stack([F[:, 1], -10.0 + tf.zeros(tf.shape(F)[0], dtype=F.dtype)], axis=1))
+            F = tf.stack([F[:, 0], tf.zeros(tf.shape(F)[0], dtype=F.dtype)], axis=1)
+            Ylabel = 1 - Y  # We need the 1 - Y, as we need to pass the *label* to SoftMax
+
+            def logistic_link(x):
+                return 1.0 / (1.0 + tf.exp(-x))
+
+            ls = gpflow.likelihoods.SoftMax(2)
+            ls.num_monte_carlo_points = 10000000
+            ls.compile()
+            lb = gpflow.likelihoods.Bernoulli(invlink=logistic_link)
+            lb.num_gauss_hermite_points = 50
+            lb.compile()
+
+            ls_cm = sess.run(ls.conditional_mean(F), feed_dict=feed)[:, :1]
+            lb_cm = sess.run(lb.conditional_mean(F[:, :1]), feed_dict=feed)
+            ls_cv = sess.run(ls.conditional_variance(F), feed_dict=feed)[:, :1]
+            lb_cv = sess.run(lb.conditional_variance(F[:, :1]), feed_dict=feed)
+            ls_lp = sess.run(ls.logp(F, Ylabel), feed_dict=feed)
+            lb_lp = sess.run(lb.logp(F[:, :1], Y), feed_dict=feed)
+
+            assert_allclose(ls_cm, lb_cm)
+            assert_allclose(ls_cv, lb_cv)
+            assert_allclose(ls_lp, lb_lp)
+
+            ls_pm, ls_pv = sess.run(ls.predict_mean_and_var(F, Fvar), feed_dict=feed)
+            lb_pm, lb_pv = sess.run(lb.predict_mean_and_var(F[:, :1], Fvar[:, :1]), feed_dict=feed)
+
+            assert_allclose(ls_pm[:, 0, None], lb_pm, rtol=1e-3)
+            assert_allclose(ls_pv[:, 0, None], lb_pv, rtol=1e-3)
+
+            ls_ve = sess.run(ls.variational_expectations(F, Fvar, Ylabel), feed_dict=feed)
+            lb_ve = sess.run(lb.variational_expectations(F[:, :1], Fvar[:, :1], Y), feed_dict=feed)
+
+            assert_allclose(ls_ve[:, 0, None], lb_ve, rtol=1e-3)
 
 
 class TestRobustMaxMulticlass(GPflowTestCase):
     """
     Some specialized tests to the multiclass likelihood with RobustMax inverse link function.
     """
+
     def setUp(self):
         self.test_graph = tf.Graph()
 
@@ -208,19 +355,19 @@ class TestRobustMaxMulticlass(GPflowTestCase):
             pred = session.run(l.predict_density(F, F, Y), feed_dict=feed)
             variational_expectations = session.run(
                 l.variational_expectations(F, F, Y), feed_dict=feed)
-            expected_mu = (1./nClasses * (1. - epsilon) + (1. - 1. / nClasses) *\
+            expected_mu = (1. / nClasses * (1. - epsilon) + (1. - 1. / nClasses) * \
                            epsilon / (nClasses - 1)) * np.ones((nPoints, 1))
 
-            self.assertTrue(np.allclose(mu, expected_mu, tolerance, tolerance))
+            self.assertTrue(np.allclose(mu, expected_mu, tolerance,
+                                        tolerance))  # assert_allclose() would complain about shape mismatch
             expected_log_denisty = np.log(expected_mu)
             self.assertTrue(np.allclose(pred, expected_log_denisty, 1e-3, 1e-3))
-            validation_variational_expectation = 1./nClasses * np.log(1. - epsilon) + \
-                (1. - 1./nClasses) * np.log(epsilon / (nClasses - 1))
-            self.assertTrue(
-                np.allclose(
-                    variational_expectations,
-                    np.ones((nPoints, 1)) * validation_variational_expectation,
-                    tolerance, tolerance))
+            validation_variational_expectation = 1. / nClasses * np.log(1. - epsilon) + \
+                                                 (1. - 1. / nClasses) * np.log(epsilon / (nClasses - 1))
+            assert_allclose(
+                variational_expectations,
+                np.ones((nPoints, 1)) * validation_variational_expectation,
+                tolerance, tolerance)
 
     def testPredictDensity(self):
         tol = 1e-4
@@ -229,7 +376,7 @@ class TestRobustMaxMulticlass(GPflowTestCase):
 
         class MockRobustMax(gpflow.likelihoods.RobustMax):
             def prob_is_largest(self, Y, Fmu, Fvar, gh_x, gh_w):
-                return tf.ones((num_points, 1)) * mock_prob
+                return tf.ones((num_points, 1), dtype=settings.float_type) * mock_prob
 
         with self.test_context() as session:
             epsilon = 0.231
@@ -250,13 +397,33 @@ class TestRobustMaxMulticlass(GPflowTestCase):
             # ^^^ evaluated on calculator:
             # log((1-\epsilon) * 0.73 + (1-0.73) * \epsilon/(num_classes -1))
 
-            self.assertTrue(np.allclose(pred, expected_prediction, tol, tol))
+            assert_allclose(pred, expected_prediction, tol, tol)
+
+    def testEpsK1Changes(self):
+        """
+        Checks that eps K1 changes when epsilon changes. This used to not happen and had to be manually changed.
+        """
+        with self.test_context() as session:
+            initial_eps = 1e-3
+            num_classes = 5
+            rm = gpflow.likelihoods.RobustMax(num_classes, initial_eps)
+
+            expected_eps_k1 = initial_eps / (num_classes - 1.)
+            actual_eps_k1 = session.run(rm._eps_K1)
+            self.assertAlmostEqual(expected_eps_k1, actual_eps_k1)
+
+            new_eps = 0.412
+            rm.epsilon.assign(new_eps, session=session)
+            expected_eps_k2 = new_eps / (num_classes - 1.)
+            actual_eps_k2 = session.run(rm._eps_K1)
+            self.assertAlmostEqual(expected_eps_k2, actual_eps_k2)
 
 
 class TestMulticlassIndexFix(GPflowTestCase):
     """
     A regression test for a bug in multiclass likelihood.
     """
+
     def testA(self):
         with self.test_context():
             mu = tf.placeholder(settings.float_type)
@@ -272,17 +439,18 @@ class TestSwitchedLikelihood(GPflowTestCase):
     SwitchedLikelihood is separately tested here.
     Here, we make sure the partition-stitch works fine.
     """
+
     def setUp(self):
         self.test_graph = tf.Graph()
 
         with self.test_context():
             rng = np.random.RandomState(1)
-            self.Y_list = [rng.randn(3, 2),  rng.randn(4, 2),  rng.randn(5, 2)]
-            self.F_list = [rng.randn(3, 2),  rng.randn(4, 2),  rng.randn(5, 2)]
-            self.Fvar_list = [np.exp(rng.randn(3, 2)),  np.exp(rng.randn(4, 2)),
+            self.Y_list = [rng.randn(3, 2), rng.randn(4, 2), rng.randn(5, 2)]
+            self.F_list = [rng.randn(3, 2), rng.randn(4, 2), rng.randn(5, 2)]
+            self.Fvar_list = [np.exp(rng.randn(3, 2)), np.exp(rng.randn(4, 2)),
                               np.exp(rng.randn(5, 2))]
-            self.Y_label = [np.ones((3, 1))*0, np.ones((4, 1))*1, np.ones((5, 1))*2]
-            self.Y_perm = list(range(3+4+5))
+            self.Y_label = [np.ones((3, 1)) * 0, np.ones((4, 1)) * 1, np.ones((5, 1)) * 2]
+            self.Y_perm = list(range(3 + 4 + 5))
             rng.shuffle(self.Y_perm)
 
             # shuffle the original data
@@ -307,7 +475,7 @@ class TestSwitchedLikelihood(GPflowTestCase):
             rslts = []
             for lik, y, f in zip(self.likelihoods, self.Y_list, self.F_list):
                 rslts.append(session.run(lik.logp(f, y)))
-            self.assertTrue(np.allclose(switched_rslt, np.concatenate(rslts)[self.Y_perm, :]))
+            assert_allclose(switched_rslt, np.concatenate(rslts)[self.Y_perm, :])
 
     def test_predict_density(self):
         with self.test_context() as session:
@@ -322,7 +490,7 @@ class TestSwitchedLikelihood(GPflowTestCase):
                                        self.F_list,
                                        self.Fvar_list):
                 rslts.append(session.run(lik.predict_density(f, fvar, y)))
-            self.assertTrue(np.allclose(switched_rslt, np.concatenate(rslts)[self.Y_perm, :]))
+            assert_allclose(switched_rslt, np.concatenate(rslts)[self.Y_perm, :])
 
     def test_variational_expectations(self):
         # switchedlikelihood
@@ -337,7 +505,7 @@ class TestSwitchedLikelihood(GPflowTestCase):
                                        self.F_list,
                                        self.Fvar_list):
                 rslts.append(session.run(lik.variational_expectations(f, fvar, y)))
-            self.assertTrue(np.allclose(switched_rslt, np.concatenate(rslts)[self.Y_perm, :]))
+            assert_allclose(switched_rslt, np.concatenate(rslts)[self.Y_perm, :])
 
 
 class TestSwitchedLikelihoodRegression(GPflowTestCase):
@@ -347,6 +515,7 @@ class TestSwitchedLikelihoodRegression(GPflowTestCase):
     one. The final column of Y is used to index the switch. If the number of
     latent functions does not match, an exception will be raised.
     """
+
     def setUp(self):
         self.test_graph = tf.Graph()
 
